/*
 *
 * The program computes a Matrix Matrix product using 
 * the C interface to the BLAS subroutine DGEMM. 
 * Data are stored on RowMajor order (C-like).
 *
 * The code was created for hands-on based 
 * training @ the ICTP for SMR2706 
 *
 * Developed by: I. Girotto ( igirotto@ictp.it )
 *
 * 
 */

//interface for dgemm:   C = alpha*A(m,k)*B(k,n) + beta*C(m,k)
//clas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
//           m, n, k, alpha, A, k, B, n, beta, C, n);

#include <stdio.h>
#include <stdlib.h>
#include "utils.h"

int main( int argc, char *argv[] ){

#ifdef __DEBUG 
  double err = 0.0;
#endif

  // using long int for containing large integer numbers
  long int n_instructions;
  int  i, j, k;
  double start, end, start_my, stop_my;
  double *A;
  double *B;
  double *C;

#ifdef __DEBUG 
  // for debug purpose an additional matrix is allocated
  double *CBLAS;
#endif

  // Declaration and initialization of temporary variables for DGEMM call
  int  SIZE = 2048;  // m, k, k, n : respectively
  double alpha = 1.0, beta = 0.0; 

  // Allocation of the Matrices
  A = (double *) malloc( SIZE * SIZE * sizeof( double ) );
  B = (double *) malloc( SIZE * SIZE * sizeof( double ) );
  C = (double *) malloc( SIZE * SIZE * sizeof( double ) );

#ifdef __DEBUG 
  CBLAS = (double *) malloc( SIZE * SIZE * sizeof( double ) );
#endif

  // Initialization of the sample matrices, beta is equal 0 so Initializzation for Matrix C is not needed
  FillMatrix( A, SIZE, SIZE );
  FillMatrix( B, SIZE, SIZE );

  start_my = cclock();
  for( i = 0; i < SIZE; i++ ) {
    for( j = 0; j < SIZE; j++ ) {
      for( k = 0; k < SIZE; k++ ) {

	C[ i * SIZE + j ] += A[ i * SIZE + k ] * B[ k * SIZE + j ];
      }
    }
  }
  stop_my = cclock();

  #ifdef __DEBUG 
  memcpy( CBLAS, C, SIZE * SIZE * sizeof( double ) );
  #endif

  // Time measuring for the eingenvalues computation
  start = cclock();
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, 
                SIZE, SIZE, SIZE, alpha, A, SIZE, B, SIZE, beta, C, SIZE);
  end = cclock();
  
  n_instructions = (long int) SIZE * SIZE * SIZE * 2;
  fprintf( stdout, "\nTime to compute MM using self made trivial code: %.3g (seconds)\n", stop_my - start_my );
  fprintf( stdout, "\nTime to compute MM using CBLAS function DGEMM: %.3g (seconds)\n", end - start );
  fprintf( stdout, "\nPerformance measured for running self made trivial code = %.3g GFLOP/s", (double) n_instructions  / 1000000000 / ( stop_my - start_my ) );
  fprintf( stdout, "\nPerformance measured for running self CBLAS DGEMM routine = %.3g GFLOP/s \n", (double) n_instructions  / 1000000000 / ( end - start ) );

#ifdef __DEBUG
  for( i = 0; i < SIZE; i++ ) {
    for( j = 0; j < SIZE; j++ ) {
      err += CBLAS[ i * SIZE + j] - C[ i * SIZE + j];
    }
  }
  fprintf( stdout, "Error comparing the two MM = %.3g", err );
#endif
  
  // free and cleaning memory
  free(A);
  free(B);
  free(C);

#ifdef __DEBUG
  free(CBLAS);
#endif 

  return 0;

}
