PROGRAM F_DGEMM
  !
  !
  !  The program computes a Matrix Matrix product using 
  !  the C interface to the BLAS subroutine DGEMM. 
  !  Data are stored on RowMajor order (C-like).
  !
  !  The code was created for hands-on based 
  !  training @ the ICTP for SMR2706 
  !
  !  Developed by: I. Girotto ( igirotto@ictp.it )
  !
  ! 
  !

  USE FillHermitian, ONLY : fillMatrix

  IMPLICIT NONE

  INTEGER, PARAMETER :: SIZE = 2048
  INTEGER :: i, j, k

#ifdef __DEBUG
  REAL * 8 :: err = 0.0d0
  ! Definition of the  Matrix needed for debug purpose
  REAL * 8, ALLOCATABLE :: CBLAS( :, : )
#endif

  ! Definition of the  Matrix related variables types 
  REAL * 8, ALLOCATABLE :: A( : , : ), B( : , : ) , C( : , : )

  ! Temporary variables for BLAS  DGEMM call 
  CHARACTER (len =1):: transa, transb
  REAL * 8  :: alpha, beta 

  ! Temporary variables for timing the time to solution of the BLAS DGEMM function  
  REAL * 8  :: t_start, t_stop, t_start_my, t_stop_my 
  REAL * 8  :: cclock
  EXTERNAL cclock

  ! Allocation of the  Matrix
  ALLOCATE( A( SIZE, SIZE ) )
  ALLOCATE( B( SIZE, SIZE ) )
  ALLOCATE( C( SIZE, SIZE ) )

#ifdef __DEBUG
  !Allocation of a matrix for debug purpose
  ALLOCATE( CBLAS( SIZE, SIZE ) )
#endif

  ! Initialization of the sample matrices 
  CALL fillMatrix( A, SIZE, SIZE )
  CALL fillMatrix( B, SIZE, SIZE )

  ! Initialization of the BLAS DGEMM variables
  transa ='N'
  transb = 'N'
  alpha = 1.0
  beta = 0.0  

  ! Time measuring for the MM self made computation
  t_start_my = cclock( )
  do k = 1, SIZE
     do j = 1, SIZE
        do i = 1, SIZE
           C( i, j ) = C( i, j ) + A( i, k ) * B( k, j )
        end do
     end do
  end do
  t_stop_my = cclock( )

#ifdef __DEBUG 
  CBLAS = C
#endif

  ! Time measuring for the DGEMM operation
  t_start = cclock( )
  CALL dgemm( transa, transb, SIZE, SIZE, SIZE, alpha, A, SIZE, B, SIZE, beta, C, SIZE)
  t_stop = cclock( )

  WRITE ( *, 75 ) "Time to compute MM using self made trivial code: ", t_stop_my - t_start_my, " (seconds) " 
  WRITE ( *, 75 ) "Time to compute MM using BLAS function DGEMM: ", t_stop - t_start, " (seconds) "
  WRITE ( *, 75 ) 'Performance measured for running self made trivial code = ', 2.0d0 * DBLE(SIZE)**3 / 1.0d9 / (t_stop_my - t_start_my), " GFLOP/s"
  WRITE ( *, 75 ) 'Performance measured for running self made trivial code = ', 2.0d0 * DBLE(SIZE)**3 / 1.0d9 / (t_stop - t_start), " GFLOP/s"

#ifdef __DEBUG                                                                                         
  do j = 1, SIZE
     do i = 1, SIZE
        err = err + ( CBLAS( i, j ) - C( i, j ) )
     end do
  end do
  WRITE( *, * ) "Error comparing the two MM = ", err
  DEALLOCATE( CBLAS )
#endif

  ! Memory Deallocation and Cleaning 
  DEALLOCATE( A )
  DEALLOCATE( B )
  DEALLOCATE( C )
  
75 FORMAT(A,F6.2,A)

END PROGRAM F_DGEMM
