/* 
 * Program to compute eigenvalues of an Hermitian matrix using 
 * the FORTRAN interface LAPACK subroutine zheev
 *
 * Code created for hands-on based evaluation for the course
 * "P1.2 Introduction to Computer Architecture for HPC"
 * for the Master in High Performance Computing @SISSA/ICTP
 *
 * Developed by: G.P. Brandino (brandino@sissa.it) and M. Atambo (matambo@ictp.it) 
 * Reviewd by: I. Girotto ( igirotto@ictp.it )
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"


int main( int argc, char *argv[] ){

  int n = 0;
  FILE* fp;
  double start, end;
  COMPLEX *A;

  // Declaration and initialization of temporary variables for LAPACK call to ZHEEV workspace
  COMPLEX worktmp;
  COMPLEX *work ;
  int info = 0, lwork = -1, lda = 0, nlwork = 0; 
  char JOBZ = 'N', UPLO = 'U'; 
  double *w, *rwork;

  // Minimal instructions for reading the matrix size from standard input
  if( argc !=2 ){
     printf("\n");
     printf("%-20s \n\n", "Sample ZHEEV code.") ;
     printf("%-8s %-4s %-20s", "usage:", argv[0], "{matrix size}");
     return 0; 
  }

  // Allocation of the Hermitian Matrix 
  n = (int) atoi( argv[1] );
  A = (COMPLEX *) malloc( sizeof(COMPLEX) * n * n );

  // Initialization of the sample matrix
  FillMatrix( A, n );

  // Print the Initliazed Hermitian matrix
  fp = fopen( "matrix.dat", "w+b" );
  fwrite( A, sizeof(COMPLEX), n*n, fp );
  fclose(fp);
   
  /*
   * Workspace query & allocation of the workspace accordingly to the obtained lwork dimension. 
   * The FORTRAN interface is used to initialize the workspace dimensions and compute
   * the eigenvalues of the Hermitian matrix. 
   *
   */
  lda = n;
  w = (double*) malloc( n * sizeof(double) );
  rwork = (double*) malloc( 3 * ( n - 2 ) * sizeof(double) );
  zheev_( &JOBZ, &UPLO,  &n, A, &lda, w, &worktmp, &lwork, rwork, &info );
  nlwork = (int) worktmp.real;
  work = (COMPLEX*) malloc( nlwork * sizeof(COMPLEX) );

  // Time measuring for the eingenvalues computation
  start = cclock(); 	
  zheev_( &JOBZ, &UPLO, &n, A, &lda, w, work, &nlwork, rwork, &info );
  end =  cclock();

  fprintf( stdout, "\nTime to compute eigenvalues for an Hermitian matrix of dimension %d using LAPACK function ZHEEV: %.3g (seconds)\n", n, end - start );

  // Write the eigenvalues of the Hermitian matrix
  fp = fopen( "eigenvalues.dat", "w+b" );
  fwrite( w, sizeof(COMPLEX), n, fp );
  fclose(fp);
  
  free(A);
  free(work);
  free(w);
  free(rwork);

  return 0;
}
